#!/usr/bin/python

import base64
import io
import tempfile
import logging
import os
import subprocess
import sys
import traceback

# python 2 compatibility
try:
    unicode()
except NameError:
    unicode = str


def log_method(fn):
    def wrapper(cls, *args, **kwargs):
        logger.log(logging.getLevelName('TRACE'),
                   ">>> Calling {}.{}, args={}, kwargs={}".format(
                       repr(cls), repr(fn), args, kwargs))
        retval = fn(cls, *args, **kwargs)
        logger.log(logging.getLevelName('TRACE'),
                   "<<< Exiting {}.{}, retval: {}".format(
                       repr(cls), repr(fn), retval))
        return retval
    return wrapper


class XccdfResult(SystemExit):
    return_code = {
        'PASS': 101,
        'FAIL': 102,
        'ERROR': 103,
        'UNKNOWN': 104,
        'NOT_APPLICABLE': 105,
        'NOT_CHECKED': 106,
        'NOT_SELECTED': 107,
        'INFORMATIONAL': 108,
        'FIXED': 109,
    }

    def __init__(self, status, *args, **kwargs):
        code = self.return_code.get(status)
        message = None
        if code is None:
            code = 1
            message = 'Bad status {}, allowed values are {}'\
                .format(status, self.return_code.keys())
        else:
            code = os.environ.get(status, code)
        super(XccdfResult, self).__init__(*args, **kwargs)
        self.code = code
        self.status = status
        if message:
            self.message = message

    def __str__(self):
        if getattr(self, 'message', ''):
            return "{}/{} {}".format(self.status, self.code, self.message)
        else:
            return "{}/{}".format(self.status, self.code)


class TextIO_Wrapper(io.TextIOWrapper):
    def __iter__(self):
        self.seek(0)
        return super(TextIO_Wrapper, self).__iter__()

    def __next__(self):
        line = super(TextIO_Wrapper, self).__next__()
        return line.rstrip()

    # python 2 compatibility
    def next(self):
        line = super(TextIO_Wrapper, self).next()
        return line.rstrip()

    def value(self):
        pos = self.tell()
        self.seek(0)
        value = self.readline().rstrip()
        self.seek(pos)
        return value


class Artifact(object):
    def __init__(self, expect=None, negate=False):
        self._expect = expect
        self._name = None
        self._description = None
        self.negate = bool(negate)
        self.value = None

    # python 2 compatibility
    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        if self.value is None:
            self.value = self()
        return self.value != self.negate

    def __str__(self):
        if self.value is None:
            return 'NOT EVALUATED'
        else:
            if self.value != self.negate:
                value = 'PASSED'
            else:
                value = 'FAILED'

            if self.negate:
                return '~{}~'.format(value)
            return value

    @log_method
    def __call__(self, *args, **kwargs):
        logger.info("Evaluating artifact {}".format(self.name))

        if self.value is None:
            self.value = self.evaluate(*args, **kwargs)

        if self.value is None:
            raise Exception("{}.evaluate() returned None".format(repr(self)))

        logger.info("Artifact {} evaluated to {}".format(self.name, bool(self)))

        return self.value

    def __invert__(self):
        return self.clone(negate=True)

    @property
    def check(self):
        return self._expect.check

    @property
    def description(self):
        if self._description is None:
            return self.check.description
        return self._description

    @property
    def name(self):
        if self._name is None:
            name = self.check.name
        else:
            name = self._name

        if self.negate:
            return '~{}~'.format(name)
        return name

    def clone(self, negate=False):
        return Artifact(expect=self._expect, negate=negate)

    def describe(self):
        f = []

        if self.name:
            f.append(self.name)
        if self.description:
            f.append('({})'.format(self.description))

        f.append('-->')
        f.append(str(self))

        return ' '.join(f)

    def evaluate(self, *args, **kwargs):
        if self._expect is None:
            return None
        else:
            return self._expect()

    def expect(self, name, *args, **kwargs):
        return Expect.func(name)(parent=self._expect)\
            .arguments(*args, **kwargs).artifact()

    def title(self, name=None, description=None):
        if name:
            self._name = name
        if description:
            self._description = description
        return self


class Expect(object):
    __function__ = None
    __classes__ = {}

    def __init__(self, check=None, parent=None):
        self.args = []
        self.kwargs = {}
        self.name = None
        self._check = check
        self._parent = parent
        self._value = None
        self._artifact = None

    @classmethod
    def func(cls, name):
        return cls.__classes__.get(name)

    @classmethod
    def register(cls, func=None):
        if func is None:
            func = getattr(cls, '__function__', None)
        logger.log(logging.getLevelName('TRACE'),
                   "Registering Expect class {} as '{}'".format(
                       repr(cls), func))
        if func is None:
            raise Exception("Expect function name is not defined")
        if func in cls.__classes__:
            raise Exception("Expect function '{}' already registered. "
                            "To override please unregister it first."
                            .format(func))
        cls.__classes__[func] = cls

    @classmethod
    def unregister(cls, func=None):
        if func is None:
            func = getattr(cls, '__function__', None)
        logger.log(logging.getLevelName('TRACE'),
                   "Unregistering Expect class {} as {}".format(
                       repr(cls), func))
        cls.__classes__.pop(func, None)

    # python 2 compatibility
    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return self._value == True

    def __str__(self):
        if self._value is None:
            return 'NOT EVALUATED'
        else:
            return str(self._value)

    @log_method
    def __call__(self, *args, **kwargs):
        description = self.describe()
        logger.debug("Expecting {}".format(description))
        if self._parent is None:
            self.check()
            self.evaluate()
        else:
            self._parent()
            if bool(self._parent):
                self.evaluate()
            else:
                return bool(self._parent)

        logger.debug("Expect {} evaluated to {}"
                     .format(self.describe(), bool(self)))
        return bool(self)

    @property
    def check(self):
        if self._parent is None:
            return self._check
        return self._parent.check

    def arguments(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        return self

    def describe(self):
        return str(self._describe(*self.args, **self.kwargs))

    def _describe(self, *args, **kwargs):
        return ""

    @log_method
    def evaluate(self):
        if self._value is None:
            self._value = bool(self._evaluate(*self.args, **self.kwargs))
        return  self._value

    def _evaluate(self, *args, **kwargs):
        raise NotImplementedError()

    def artifact(self):
        self._artifact = Artifact(expect=self)
        return self._artifact

    def title(self, *args, **kwargs):
        self._artifact.title(*args, **kwargs)


class Check(object):
    def __init__(self, name=None, description=None):
        self.name = name
        self.description = description
        self._evaluated = False
        self.stdout = None
        logger.log(logging.getLevelName('TRACE'),
                   "Check {} added: {}".format(str(self), repr(self)))

    def __str__(self):
        if self.name is None:
            return '<name not set>'
        return str(self.name)

    @log_method
    def __call__(self, *args, **kwargs):
        if not self._evaluated:
            self.evaluate(*args, **kwargs)
            self._evaluated = True

    def title(self, name=None, description=None):
        if name:
            self.name = name
        if description:
            self.description = description
        return self

    def evaluate(self, *args, **kwargs):
        raise NotImplementedError("Check.eval() not implemented")

    def expect(self, name, *args, **kwargs):
        return Expect.func(name)(check=self)\
            .arguments(*args, **kwargs).artifact()


class ShCheck(Check):
    def __init__(self, *args, **kwargs):
        super(ShCheck, self).__init__(*args, **kwargs)
        self._cmd = None
        self._script = None
        self.stderr = None
        self.exitcode = None

    def cmd(self, value):
        self._cmd = value.strip()
        return self

    @log_method
    def evaluate(self):
        script_file = None
        cmd_args = ['/bin/bash', '-x']
        if self._cmd:
            cmd_args.extend(['-c', self._cmd])
        elif self._script:
            _, script_file = tempfile.mkstemp()
            with open(script_file, 'w') as f:
                f.write(self._script)
            cmd_args.append(script_file)

        stderr_fd, stderr_filename = tempfile.mkstemp()
        try:
            stdout = subprocess.check_output(
                cmd_args, stderr=stderr_fd)
            self.exitcode = 0
        except subprocess.CalledProcessError as e:
            self.exitcode = e.returncode
            stdout = e.output
        finally:
            if script_file:
                os.remove(script_file)

        self.stdout = TextIO_Wrapper(io.BytesIO(stdout))
        with open(stderr_filename, 'rb') as f:
            self.stderr = TextIO_Wrapper(io.BytesIO(f.read()))
        os.remove(stderr_filename)

        logger.debug('ARGS: %s', cmd_args)
        logger.debug('EXIT_CODE: %s', self.exitcode)

        for line in self.stdout:
            logger.debug('STDOUT: %s', line)

        for line in self.stderr:
            logger.debug('STDERR: %s', line)

    def script(self, value):
        self._script = value.strip()
        return self


log_level = 'DEBUG'
log_format = '%(asctime)-15s [%(levelname)-5s] %(message)s'

log_level = os.environ.get('XCCDF_VALUE_LOG_LEVEL', log_level)
log_format = os.environ.get('XCCDF_VALUE_LOG_FORMAT', log_format)

logging.addLevelName(5, 'TRACE')
logging.basicConfig(level=logging.getLevelName(log_level), format=log_format)

logger = logging.getLogger()


class ExpectValue(Expect):
    __function__ = 'value'

    def _describe(self, value):
        return "value '{}'".format(value)

    def _evaluate(self, value):
        return self.check.stdout.value() == str(value)


class ExpectExitCode(Expect):
    __function__ = 'exitcode'

    def _describe(self, value):
        return "exitcode {}".format(value)

    def _evaluate(self, value):
        return self.check.exitcode == value


ExpectExitCode.register()
ExpectValue.register()
Sh = ShCheck

if len(sys.argv) > 1:
    entrypoint = sys.argv[1]
else:
    entrypoint = None

if entrypoint is None:
    entrypoint = os.environ.get('XCCDF_VALUE_ENTRYPOINT')

remove_entrypoint = False
if entrypoint is None:
    codeblock = os.environ.get('XCCDF_VALUE_CODEBLOCK')
    if codeblock:
        _, entrypoint = tempfile.mkstemp()
        with open(entrypoint, 'w') as f:
            f.write(base64.b64decode(codeblock))
        remove_entrypoint = True

if entrypoint is None:
    raise XccdfResult('ERROR', 'Entrypoint / codeblock is not provided')

PASS_STATUS='PASS'
FAIL_STATUS='FAIL'
RESULT = None
try:
    logger.info("Processing file {}".format(entrypoint))
    pyc = compile(open(entrypoint, "rb").read(), entrypoint, 'exec')
    if remove_entrypoint:
        os.remove(entrypoint)
    exec(pyc, globals(), locals())
    if RESULT is None:
        raise XccdfResult('NOT_CHECKED', 'RESULT was not set')
    if not isinstance(RESULT, Artifact):
        raise XccdfResult('ERROR', 'RESULT should be an instance of Artifact, '
                                   'got {}'.format(repr(RESULT)))
    if bool(RESULT):
        raise XccdfResult(PASS_STATUS)
    else:
        raise XccdfResult(FAIL_STATUS)
except XccdfResult as e:
    if isinstance(RESULT, Artifact):
        print("Determinant: {}".format(RESULT.describe()))
    print("XCCDF result: {}".format(str(e)))
    raise e
except:
    traceback.print_exc()
    e = XccdfResult('ERROR', 'Unexpected error occured')
    print(str(e))
    raise e
